import { transformSync } from '@babel/core';
import jsxSyntax from '@babel/plugin-syntax-jsx';

// src/internal/util.ts
var includes = (value, search) => !!~value.indexOf(search);
var join = (parts, separator = "-") => parts.join(separator);
var tail = (array, startIndex = 1) => array.slice(startIndex);

// src/twind/parse.ts
var groupings;
var rules;
var startGrouping = (value = "") => {
  groupings.push(value);
  return "";
};
var endGrouping = (isWhitespace) => {
  groupings.length = Math.max(groupings.lastIndexOf("") + ~~isWhitespace, 0);
};
var onlyPrefixes = (s) => s && !includes("!:", s[0]);
var onlyVariants = (s) => s[0] == ":";
var addRule = (directive2, negate) => {
  rules.push({
    v: groupings.filter(onlyVariants),
    d: directive2,
    n: negate,
    i: includes(groupings, "!"),
    $: ""
  });
};
var saveRule = (buffer) => {
  const negate = buffer[0] == "-";
  if (negate) {
    buffer = tail(buffer);
  }
  const prefix = join(groupings.filter(onlyPrefixes));
  addRule(buffer == "&" ? prefix : (prefix && prefix + "-") + buffer, negate);
  return "";
};
var parseString = (token, isVariant) => {
  let buffer = "";
  for (let char, dynamic = false, position2 = 0; char = token[position2++]; ) {
    if (dynamic || char == "[") {
      buffer += char;
      dynamic = char != "]";
      continue;
    }
    switch (char) {
      case ":":
        buffer = buffer && startGrouping(":" + (token[position2] == char ? token[position2++] : "") + buffer);
        break;
      case "(":
        buffer = buffer && startGrouping(buffer);
        startGrouping();
        break;
      case "!":
        startGrouping(char);
        break;
      case ")":
      case " ":
      case "	":
      case "\n":
      case "\r":
        buffer = buffer && saveRule(buffer);
        endGrouping(char !== ")");
        break;
      default:
        buffer += char;
    }
  }
  if (buffer) {
    if (isVariant) {
      startGrouping(":" + buffer);
    } else if (buffer.slice(-1) == "-") {
      startGrouping(buffer.slice(0, -1));
    } else {
      saveRule(buffer);
    }
  }
};
var parseGroupedToken = (token) => {
  startGrouping();
  parseToken(token);
  endGrouping();
};
var parseGroup = (key, token) => {
  if (token) {
    startGrouping();
    const isVariant = includes("tbu", (typeof token)[1]);
    parseString(key, isVariant);
    if (isVariant) {
      parseGroupedToken(token);
    }
    endGrouping();
  }
};
var parseToken = (token) => {
  switch (typeof token) {
    case "string":
      parseString(token);
      break;
    case "function":
      addRule(token);
      break;
    case "object":
      if (Array.isArray(token)) {
        token.forEach(parseGroupedToken);
      } else if (token) {
        Object.keys(token).forEach((key) => {
          parseGroup(key, token[key]);
        });
      }
  }
};
var staticsCaches = new WeakMap();
var buildStatics = (strings) => {
  let statics = staticsCaches.get(strings);
  if (!statics) {
    let slowModeIndex = NaN;
    let buffer = "";
    statics = strings.map((token, index) => {
      if (slowModeIndex !== slowModeIndex && (token.slice(-1) == "[" || includes(":-(", (strings[index + 1] || "")[0]))) {
        slowModeIndex = index;
      }
      if (index >= slowModeIndex) {
        return (interpolation) => {
          if (index == slowModeIndex) {
            buffer = "";
          }
          buffer += token;
          if (includes("rg", (typeof interpolation)[5])) {
            buffer += interpolation;
          } else if (interpolation) {
            parseString(buffer);
            buffer = "";
            parseToken(interpolation);
          }
          if (index == strings.length - 1) {
            parseString(buffer);
          }
        };
      }
      const staticRules = rules = [];
      parseString(token);
      const activeGroupings = [...groupings];
      rules = [];
      return (interpolation) => {
        rules.push(...staticRules);
        groupings = [...activeGroupings];
        if (interpolation) {
          parseToken(interpolation);
        }
      };
    });
    staticsCaches.set(strings, statics);
  }
  return statics;
};
var parse = (tokens) => {
  groupings = [];
  rules = [];
  if (Array.isArray(tokens[0]) && Array.isArray(tokens[0].raw)) {
    buildStatics(tokens[0]).forEach((apply2, index) => apply2(tokens[index + 1]));
  } else {
    parseToken(tokens);
  }
  return rules;
};
var stringifyVariant = (selector, variant) => selector + (variant[1] == ":" ? tail(variant, 2) + ":" : tail(variant)) + ":";
var stringifyRule = (rule, directive2 = rule.d) => typeof directive2 == "function" ? "" : rule.v.reduce(stringifyVariant, "") + (rule.i ? "!" : "") + (rule.n ? "-" : "") + directive2;

// src/twind/expand.ts
var expandGroups = (classNames) => parse(classNames).map((rule) => stringifyRule(rule)).join(" ");

const {
  default: jsx
} = jsxSyntax;
/**
 * @param {import('@babel/core')} babel
 * @returns {import('@babel/core').PluginObj}
 */

function tailwindGroupingPlugin$1({
  types: t
}) {
  function hasParenthesis(value) {
    return value.includes('(') || value.includes(')');
  }

  const jsxExpressionContainerVisitor = {
    StringLiteral(path) {
      if (!hasParenthesis(path.node.value)) return;
      path.node.value = expandGroups(path.node.value);
    },

    TemplateElement(path) {
      var _path$node$value$raw$, _path$node$value$raw$2;

      if (!hasParenthesis(path.node.value.raw)) return; // If the numbers of left & right are not equal, the group
      // spans across a conditional and we need to bring in the variant or directive
      //
      // class={`text(black {value ? 'right' : 'left'} 3xl)`}
      //         ^^^^^^^^^^                            ^^^^

      if (((_path$node$value$raw$ = path.node.value.raw.match(/\(/g)) == null ? void 0 : _path$node$value$raw$.length) !== ((_path$node$value$raw$2 = path.node.value.raw.match(/\)/g)) == null ? void 0 : _path$node$value$raw$2.length)) {
        const lastGroup = path.node.value.raw.match(/([^\s]+)\((?:[^\(]+)?$/)[1]; // Add a `)` and expand (`flex(&`) or just convert `(` to a hyphen (`flex-`)

        path.node.value.raw = (/\(\s?$/.test(path.node.value.raw) ? `${expandGroups(path.node.value.raw)} ${lastGroup}-` : `${expandGroups(path.node.value.raw + ')')} ${lastGroup}-`).trimStart();
        const nextSibling = path.getNextSibling().node;

        if (t.isTemplateElement(nextSibling)) {
          const splinteredGroup = nextSibling.value.raw;
          path.getNextSibling().node.value.raw = /\S\)/.test(splinteredGroup) ? ` ${expandGroups(`${lastGroup}(${splinteredGroup}`)}` : '';
        }
      } else {
        let ending = '';

        if (/(-|\s)$/.test(path.node.value.raw)) {
          const lastVal = path.node.value.raw.lastIndexOf(' ');
          const length = path.node.value.raw.length;
          ending = path.node.value.raw.slice(lastVal - length);
        }

        let start = '';

        if (/^\s/.test(path.node.value.raw)) {
          const firstVal = path.node.value.raw.indexOf(' ') + 1;
          start = path.node.value.raw.slice(0, firstVal);
        }

        path.node.value.raw = start + expandGroups(path.node.value.raw) + ending;
      }
    }

  };
  return {
    name: 'tailwind-grouping',
    inherits: jsx,
    visitor: {
      JSXAttribute(path) {
        // @ts-ignore
        if (!/class(?:Name)?/.test(path.node.name.name) || !path.node.value) return;

        if (t.isStringLiteral(path.node.value)) {
          if (!hasParenthesis(path.node.value.value)) return;
          path.node.value.value = expandGroups(path.node.value.value);
        } else if (t.isJSXExpressionContainer(path.node.value)) {
          path.traverse(jsxExpressionContainerVisitor);
        }
      }

    }
  };
}

/** @typedef {import('@babel/core').BabelFileResult} BabelFileResult */

/**
 * @returns {{
 *   name: 'tailwind-grouping',
 *   transform: (code: string, id: string) => { code?: BabelFileResult['code'], map: BabelFileResult['map'] } | void
 * }}
 */

function tailwindGroupingPlugin() {
  return {
    name: 'tailwind-grouping',

    transform(code, id) {
      // Skip non-user code
      if (id[0] === '\0' || id.startsWith('npm/')) return; // Skip if non-JS(X)/TSX files, and files that do not contain JSX

      if (!/\.(?:jsx?|tsx)$/.test(id) || !/<[a-zA-Z$_][\w.:-]*[^>]*>/.test(code)) return;
      const isTSX = /.tsx$/.test(id);
      const result = transformSync(code, {
        plugins: [isTSX && ['@babel/plugin-syntax-typescript', {
          isTSX
        }], tailwindGroupingPlugin$1].filter(Boolean)
      });
      return {
        code: result.code,
        map: result.map
      };
    }

  };
}

export { tailwindGroupingPlugin as default };
